from functools import cached_property


class BaseChannel:
    """
    Base Channel class to be implemented by
    the client.
    """

    def send(self, data, ep_id):
        """
        Send data to endpoint ep_id. Implement this to
        send data contained in the data byte array, to
        the endpoint with id ep_id
        Arguments:
            data: a bytearray containing the data to
                  be sent
            ep_id: an integer representing the endpoint
                   ID to send to
        """
        raise NotImplementedError

    def recv(self, ep_id, timeout=0.1):
        """
        Receive data from endpoint ep_id. Implement this
        to receive data from endpoint with id ep_id. The
        implementation should timeout after a period equal
        to timeout has elapsed.
        Arguments:
            ep_id: an integer representing the endpoint ID
                   to listen to for data
            timeout: an integer indicating a timeout for
                     receiving any data
        Returns:
            A bytearray containing the received data
        """
        raise NotImplementedError

    @cached_property
    def max_ep_id(self):
        """
        Get the max value that the endpoint can get for
        this channel.
        Returns:
            An integer representing the max endpoint ID
            value
        """
        raise NotImplementedError

    @cached_property
    def max_packet_size(self):
        """
        Get the max size in bytes of the packet that this
        channel can transmit/receive.
        Returns:
            An integer representing the max packet size
        """
        raise NotImplementedError

    @cached_property
    def serializer(self):
        """
        Return a data codec appropriate for this channel.
        Returns:
            The data codec instance
        """
        raise NotImplementedError



from avlos import get_registry
from avlos.mixins.comm_node import CommNode
from avlos.mixins.named_node import NamedNode
from avlos.mixins.meta_node import MetaNode
from avlos.mixins.impex_node import ImpexNode
from avlos.mixins.func_attr_node import FuncAttrNode


class RemoteAttribute(CommNode, NamedNode, MetaNode, ImpexNode, FuncAttrNode):
    """
    Remote Endpoint with a value, parent and a comms channel
    """

    def __init__(
        self,
        name,
        summary,
        dtype,
        func_attr=None,
        getter_name=None,
        setter_name=None,
        unit=None,
        rst_target=None,
        ep_id=-1,
        meta={},
    ):
        CommNode.__init__(self)
        NamedNode.__init__(self, name)
        MetaNode.__init__(self, meta_dict=meta)
        FuncAttrNode.__init__(self, func_attr)
        self.summary = summary
        self.dtype = dtype
        self.unit = unit
        self.getter_name = getter_name
        self.setter_name = setter_name
        self.rst_target = rst_target
        self.ep_id = ep_id

    def get_value(self):
        """
        Retrieve the current value from the remote endpoint.
        Sends a request over the channel, receives the response, and deserializes it.
        Applies unit conversion if the attribute has an associated unit.

        Returns:
            The deserialized value, optionally with unit attached
        """
        assert self.getter_name, "No getter function available"
        self.channel.send([], self.ep_id)
        data = self.channel.recv(self.ep_id)
        value, *_ = self.channel.serializer.deserialize(data, self.dtype)
        try:
            return value * self.unit
        except TypeError:
            return value

    def set_value(self, __value):
        """
        Set a new value on the remote endpoint.
        Converts units if necessary, serializes the value, and sends it over the channel.

        Args:
            __value: The value to set, optionally with units attached
        """
        assert self.setter_name, "No setter function available"
        try:
            __value = __value.to(self.unit).magnitude
        except AttributeError:
            pass
        data = self.channel.serializer.serialize([__value], self.dtype)
        self.channel.send(data, self.ep_id)

    def set_value_with_string(self, __str_value):
        """
        Set a value from a string representation, parsing it with the unit registry.

        Args:
            __str_value: String representation of the value to set
        """
        self.set_value(get_registry()(__str_value))

    def str_dump(self):
        """
        Generate a formatted string representation of the attribute and its current value.

        Returns:
            A formatted string showing the attribute name, data type, and current value
        """
        value = self.get_value()
        if isinstance(value, (int, float)):
            format_str = "{0} [{1}]: {2:.6g}"
        else:
            format_str = "{0} [{1}]: {2}"
        return format_str.format(
            self.name,
            self.dtype.nickname,
            value
        )



import json
import pint
from avlos import get_registry


class AvlosEncoder(json.JSONEncoder):
    """
    JSON Encoder that serializes pint Quantity
    objects to strings
    """

    def default(self, o):
        if isinstance(o, pint.Quantity):
            return str(o)
        else:
            return super().default(o)



import enum
from marshmallow import fields, ValidationError


class EnumField(fields.Field):
    """
    Marshmallow Field that serializes to a string
    and deserializes to an enum.
    """

    def _serialize(self, value, attr, obj, **kwargs):
        return NotImplementedError

    def _deserialize(self, value, attr, data, **kwargs):
        try:
            assert len(value) > 0
            return enum.IntEnum(attr, value, start=0)
        except ValueError as error:
            raise ValidationError("Invalid enum list.") from error
        except AssertionError as error:
            raise ValidationError("Empty enum list.") from error



import json
import hashlib
from avlos.definitions import RootNodeSchema
from avlos.counter import make_counter


def deserialize(device_description):
    make_counter()
    device_schema = RootNodeSchema()
    device_obj = device_schema.load(device_description)
    dev_desc = json.dumps(device_description)
    device_obj.hash_string = hash_string_from_string(dev_desc)
    device_obj.hash_uint32 = hash_int_from_string(dev_desc)
    return device_obj


def hash_string_from_string(input_string):
    return hex(hash_int_from_string(input_string))


def hash_int_from_string(input_string):
    return int.from_bytes(hashlib.sha256(input_string.encode()).digest()[:4], "little")



import os
from jinja2 import Environment, PackageLoader, select_autoescape
from avlos.generators.filters import avlos_endpoints

env = Environment(loader=PackageLoader("avlos"), autoescape=select_autoescape())


def process(instance, config):
    env.filters["endpoints"] = avlos_endpoints

    template = env.get_template("docs.rst.jinja")
    os.makedirs(os.path.dirname(config["paths"]["output_file"]), exist_ok=True)
    with open(config["paths"]["output_file"], "w") as output_file:
        print(
            template.render(instance=instance),
            file=output_file,
        )



import yaml
import importlib.resources
from avlos.deserializer import deserialize
from avlos.unit_field import get_registry
import unittest
from tests.dummy_channel import DummyChannel

_reg = get_registry()


class TestRemoteObjects(unittest.TestCase):
    def test_read_remote_properties(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath("definition/good_device.yaml")
        )
        with open(def_path_str) as device_description:
            obj = deserialize(yaml.safe_load(device_description))
            obj._channel = DummyChannel()

            self.assertEqual(0, obj.Vbus)
            obj._channel.set_value(12.0)
            self.assertEqual(12.0 * _reg("volt"), obj.Vbus)
            obj._channel.set_value(24.0)
            self.assertEqual(24.0 * _reg("volt"), obj.Vbus)

            obj._channel.set_value(0.1)
            self.assertEqual(0.1 * _reg("ohm"), obj.motor.R)

            obj.channel.set_value("lalala")
            self.assertEqual(obj.nickname, "lalala")

            obj.controller.set_pos_vel_setpoints(0, 0)

            obj._channel.write_on()
            obj.nickname = "other"
            obj._channel.write_off()
            self.assertEqual(obj.nickname, "other")

    def test_remote_enum_read(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath("definition/good_device.yaml")
        )
        with open(def_path_str) as device_description:
            obj = deserialize(yaml.safe_load(device_description))
            obj._channel = DummyChannel()
            modes = obj.controller.remote_attributes["mode"].options
            obj._channel.set_value(0)
            self.assertEqual(obj.controller.mode, modes.IDLE)
            obj._channel.set_value(1)
            self.assertEqual(obj.controller.mode, modes.CLOSED_LOOP)

    def test_remote_enum_write(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath("definition/good_device.yaml")
        )
        with open(def_path_str) as device_description:
            obj = deserialize(yaml.safe_load(device_description))
            obj._channel = DummyChannel()
            modes = obj.controller.remote_attributes["mode"].options
            obj._channel.write_on()

            # Test setting the value using an integer
            obj.controller.mode = 0
            self.assertEqual(obj._channel.value, 0)

            # Test setting the value using another integer
            obj.controller.mode = 1
            self.assertEqual(obj._channel.value, 1)

            # Test setting the value using an enum member
            obj.controller.mode = modes.IDLE
            self.assertEqual(obj._channel.value, 0)

            # Test setting the value using another enum member
            obj.controller.mode = modes.CLOSED_LOOP
            self.assertEqual(obj._channel.value, 1)

            # Test setting the value using a string corresponding to an enum member
            obj.controller.mode = "IDLE"
            self.assertEqual(obj._channel.value, 0)

            # Test setting the value using another string corresponding to an enum member
            obj.controller.mode = "CLOSED_LOOP"
            self.assertEqual(obj._channel.value, 1)

    def test_remote_function_call(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath("definition/good_device.yaml")
        )
        with open(def_path_str) as device_description:
            obj = deserialize(yaml.safe_load(device_description))
            obj._channel = DummyChannel()

        self.assertEqual(0, obj.controller.set_pos_vel_setpoints(1, 2))
        obj._channel.set_value(100.0)
        self.assertEqual(100 * _reg("tick"), obj.controller.set_pos_vel_setpoints(0, 0))

    def test_remote_function_call_w_units(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath("definition/good_device.yaml")
        )
        with open(def_path_str) as device_description:
            obj = deserialize(yaml.safe_load(device_description))
            obj._channel = DummyChannel()

        obj._channel.write_on()
        obj.move_to(1 * _reg("turn"))
        self.assertEqual(8192, obj._channel.value)
        obj.move_to(-100)
        self.assertEqual(-100, obj._channel.value)
        obj._channel.write_off()

    def test_non_existent_remote_attributes_fail(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath("definition/good_device.yaml")
        )
        with open(def_path_str) as device_description:
            obj = deserialize(yaml.safe_load(device_description))
            obj._channel = DummyChannel()
            with self.assertRaises(AttributeError):
                val = obj.foo
                print(val)
            with self.assertRaises(AttributeError):
                val = obj.controller.bar
                print(val)

    def test_meta_dictionary(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath("definition/good_device.yaml")
        )
        with open(def_path_str) as device_description:
            obj = deserialize(yaml.safe_load(device_description))
            self.assertEqual(1, len(obj.errors.meta))
            self.assertEqual("ok", obj.errors.meta["lalala"])
            self.assertEqual(1, len(obj.reset.meta))
            self.assertEqual(True, obj.reset.meta["reload_data"])
            self.assertEqual(0, len(obj.sn.meta))
            with self.assertRaises(KeyError):
                d = obj.sn.meta["reload_data"]



from collections import OrderedDict
from marshmallow import (
    Schema,
    fields,
    post_load,
    validates_schema,
    ValidationError,
)
from avlos.unit_field import UnitField
from avlos.bitmask_field import BitmaskField
from avlos.enum_field import EnumField
from avlos.counter import get_counter
from avlos.datatypes import DataTypeField
from avlos.mixins.comm_node import CommNode
from avlos.mixins.named_node import NamedNode
from avlos.mixins.impex_node import ImpexNode
from avlos.definitions import (
    RemoteAttribute,
    RemoteFunction,
    RemoteArgumentSchema,
    RemoteBitmask,
    RemoteEnum,
)


class RemoteNode(CommNode, NamedNode, ImpexNode):
    """
    Remote node with parent, children and a comms channel
    """

    def __init__(self, remote_attributes, name, summary=None):
        """
        Initialize a remote node with attributes organized as an ordered dictionary.

        Args:
            remote_attributes: List of remote attributes/functions/nodes to be accessed by name
            name: The name identifier for this node
            summary: Optional description of the node's purpose
        """
        od = OrderedDict()
        for attrib in remote_attributes:
            od[attrib.name] = attrib
        super().__setattr__("remote_attributes", od)
        CommNode.__init__(self)
        NamedNode.__init__(self, name)
        self.summary = summary

    def __getattr__(self, __name):
        """
        Enable dynamic attribute access for remote attributes.
        Attempts to call get_value() if the attribute supports it, otherwise returns the attribute itself.

        Args:
            __name: Name of the attribute to access

        Returns:
            The value of the remote attribute, or the attribute object itself

        Raises:
            AttributeError: If the attribute name doesn't exist in remote_attributes
        """
        try:
            attr = self.remote_attributes[__name]
            try:
                return attr.get_value()
            except AttributeError:
                return attr
        except KeyError:
            raise AttributeError(__name)

    def __setattr__(self, __name, __value):
        """
        Enable dynamic attribute setting for remote attributes.
        Delegates to the attribute's set_value() method if it's a remote attribute,
        otherwise falls back to standard attribute setting.

        Args:
            __name: Name of the attribute to set
            __value: Value to set the attribute to
        """
        try:
            attr = self.remote_attributes[__name]
            return attr.set_value(__value)
        except KeyError:
            super().__setattr__(__name, __value)

    def export_flags(self, namespace):
        """
        Recurse children and export bitmask to
        indicated namespace
        """
        for child in self.remote_attributes:
            try:
                child.export_flags(namespace)
            except AttributeError:
                pass

    def str_dump(self, indent, depth):
        """
        Generate a formatted string representation of the node and its children.
        Recursively traverses child nodes up to the specified depth.

        Args:
            indent: String used for indentation at the current level
            depth: Maximum recursion depth for displaying nested nodes

        Returns:
            Formatted multi-line string representation of the node hierarchy
        """
        if depth <= 0:
            return "..."
        lines = []
        for key, val in self.remote_attributes.items():
            if isinstance(val, RemoteNode):
                val_str = (
                    indent
                    + key
                    + (": " if depth == 1 else ":\n")
                    + val.str_dump(indent + "  ", depth - 1)
                )
            else:
                val_str = indent + val.str_dump()
            lines.append(val_str)
        return "\n".join(lines)

    def __str__(self):
        return self.str_dump("", depth=2)

    def __repr__(self):
        return self.__str__()

    def __dir__(self):
        return self.remote_attributes.keys()


class RemoteNodeSchema(Schema):
    """
    Custom Marshmallow schema for generating RemoteNode,
    RemoteAttribute, RemoteBitmask and RemoteFunction classes
    """

    name = fields.String(
        required=True, error_messages={"required": "Name is required."}
    )
    summary = fields.String()
    remote_attributes = fields.List(fields.Nested(lambda: RemoteNodeSchema()))
    dtype = DataTypeField()
    flags = BitmaskField()
    options = EnumField()
    unit = UnitField()
    func_attr = fields.String(dump_default=None)
    getter_name = fields.String()
    setter_name = fields.String()
    caller_name = fields.String()
    arguments = fields.List(fields.Nested(lambda: RemoteArgumentSchema()))
    export = fields.Bool(dump_default=False)
    rst_target = fields.String()
    ep_id = fields.Integer(dump_default=-1)
    meta = fields.Dict(dump_default={})

    def __init__(self, *args, **kwargs):
        """
        Initialize the schema with a counter for assigning unique endpoint IDs.
        """
        self.counter = get_counter()
        super().__init__(*args, **kwargs)

    @post_load
    def make_remote_node(self, data, **kwargs):
        """
        Decide on which type of object to instantiate
        based on the initialization data available.
        Overrides the post_load hook.
        """
        if "remote_attributes" in data:
            node = RemoteNode(**data)
            for child in node.remote_attributes.values():
                child._parent = node
            return node
        elif "caller_name" in data:
            data["ep_id"] = self.counter.next()
            return RemoteFunction(**data)
        elif "dtype" in data:
            data["ep_id"] = self.counter.next()
            return RemoteAttribute(**data)
        elif "flags" in data:
            data["ep_id"] = self.counter.next()
            return RemoteBitmask(**data)
        elif "options" in data:
            data["ep_id"] = self.counter.next()
            return RemoteEnum(**data)

    @validates_schema
    def validate_schema(self, data, **kwargs):
        """
        Validate that the schema data meets structural requirements.
        Ensures nodes have either child attributes or endpoint functionality,
        and that endpoints have appropriate data type definitions.

        Raises:
            ValidationError: If validation rules are violated
        """
        if (
            "remote_attributes" not in data
            and "getter_name" not in data
            and "setter_name" not in data
            and "caller_name" not in data
        ):
            raise ValidationError(
                "Either a getter, setter, caller or remote attributes list is required"
            )
        if "getter_name" in data and "setter_name" in data and "caller_name" in data:
            raise ValidationError(
                "A getter, setter, and caller cannot coexist in a single endpoint"
            )
        if (
            ("getter_name" in data or "setter_name" in data or "caller_name" in data)
            and "dtype" not in data
            and "flags" not in data
            and "options" not in data
        ):
            raise ValidationError("Data type, flags or options field is required")



from avlos.generators.filters import as_include, file_from_path, capitalize_first
import unittest


class TestFunctions(unittest.TestCase):
    def test_file_from_path_filter(self):
        self.assertEqual("test_functions.py", file_from_path(__file__))

    def test_jinja_as_include_filter(self):
        self.assertEqual("<test.h>", as_include("test.h"))
        self.assertEqual("<test.h>", as_include("<test.h>"))
        self.assertEqual('"test.h"', as_include('"test.h"'))

    def test_capitalize_first(self):
        self.assertEqual(capitalize_first("bob Meadow"), "Bob Meadow")



import json
import yaml
import importlib.resources
from avlos.deserializer import deserialize
from avlos.unit_field import get_registry
from avlos.json_codec import AvlosEncoder
import unittest
from tests.dummy_channel import DummyChannel


class TestImpex(unittest.TestCase):
    def test_import_export_root_object(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath("definition/good_device.yaml")
        )
        with open(def_path_str) as device_description:
            obj = deserialize(yaml.safe_load(device_description))
            obj._channel = DummyChannel()
            obj._channel.set_value(0)
            func = obj.motor.remote_attributes["R"]
            root = func.root
            self.assertEqual(obj, root)
            values = root.export_values()
            json_string = json.dumps(values, cls=AvlosEncoder)
            imported_values = json.loads(json_string)
            root.import_values(imported_values)



from enum import Enum
from marshmallow import fields, ValidationError


class DataType(Enum):
    """
    Enum that represents a datatype. Includes
    various convenience functions
    """

    VOID = 0
    BOOL = 1
    UINT8 = 2
    INT8 = 3
    UINT16 = 4
    INT16 = 5
    UINT32 = 6
    INT32 = 7
    UINT64 = 8
    INT64 = 9
    FLOAT = 10
    DOUBLE = 11
    STR = 12

    @property
    def nickname(self):
        return self.name.lower()

    @property
    def c_name(self):
        return c_type_map[self]

    @property
    def size(self):
        return datatype_sizes[self]

    @property
    def is_void(self):
        return DataType.VOID == self

    def from_string(self, str_value):
        if self == DataType.VOID:
            return None
        else:
            return py_type_map[self](str_value)


c_type_map = {
    DataType.VOID: "void",
    DataType.BOOL: "bool",
    DataType.INT8: "int8_t",
    DataType.UINT8: "uint8_t",
    DataType.INT16: "int16_t",
    DataType.UINT16: "uint16_t",
    DataType.INT32: "int32_t",
    DataType.UINT32: "uint32_t",
    DataType.FLOAT: "float",
    DataType.DOUBLE: "double",
    DataType.STR: "char[]",
}

py_type_map = {
    DataType.VOID: None,
    DataType.BOOL: bool,
    DataType.INT8: int,
    DataType.UINT8: int,
    DataType.INT16: int,
    DataType.UINT16: int,
    DataType.INT32: int,
    DataType.UINT32: int,
    DataType.FLOAT: float,
    DataType.DOUBLE: float,
    DataType.STR: str,
}

datatype_sizes = {
    DataType.VOID: 0,
    DataType.BOOL: 1,
    DataType.INT8: 1,
    DataType.UINT8: 1,
    DataType.INT16: 2,
    DataType.UINT16: 2,
    DataType.INT32: 4,
    DataType.UINT32: 4,
    DataType.FLOAT: 4,
    DataType.DOUBLE: 8,
    DataType.STR: -1,
}


datatype_names = {
    "void": DataType.VOID,
    "bool": DataType.BOOL,
    "int8": DataType.INT8,
    "uint8": DataType.UINT8,
    "int16": DataType.INT16,
    "uint16": DataType.UINT16,
    "int32": DataType.INT32,
    "uint32": DataType.UINT32,
    "float": DataType.FLOAT,
    "double": DataType.DOUBLE,
    "string": DataType.STR,
}


class DataTypeField(fields.Field):
    """
    Marshmallow Field that serializes to a string and
    deserializes to a member of the DataType enum.
    """

    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return ""
        for k, v in datatype_names.iteritems():
            if v == value:
                return k

    def _deserialize(self, value, attr, data, **kwargs):
        try:
            return datatype_names[value]
        except KeyError as error:
            raise ValidationError("Invalid DataType.") from error



from avlos.unit_field import get_registry

try:
    from avlos._version import __version__
except ImportError:
    # Package is not installed, version will be determined from git
    try:
        from setuptools_scm import get_version
        __version__ = get_version(root='..', relative_to=__file__)
    except (ImportError, LookupError):
        __version__ = "unknown"

__all__ = ["get_registry", "__version__"]


from functools import cached_property


class CommNode:
    def __init__(self):
        self._parent = None
        self._channel = None

    @cached_property
    def parent(self):
        return self._parent

    @cached_property
    def root(self):
        return self._parent.root

    @cached_property
    def channel(self):
        try:
            return self.parent.channel
        except AttributeError:
            return self._channel

    def set_channel(self, channel):
        self._channel = channel



from avlos.definitions.remote_attribute import RemoteAttribute
from avlos.definitions.remote_function import RemoteFunction, RemoteArgument, RemoteArgumentSchema
from avlos.definitions.remote_bitmask import RemoteBitmask
from avlos.definitions.remote_enum import RemoteEnum
from avlos.definitions.remote_node import RemoteNode, RemoteNodeSchema
from avlos.definitions.remote_root_node import RootNode, RootNodeSchema



from marshmallow import fields, ValidationError
import pint

_registry = None


def get_registry():
    """
    Get or create a Pint unit registry with custom units
    """
    global _registry
    if not _registry:
        _registry = pint.UnitRegistry(autoconvert_offset_to_baseunit=True)
        _registry.define("tick = turn / 8192")
    return _registry


class UnitField(fields.Field):
    """
    Marshmallow Field that serializes to a string
    and deserializes to a Pint unit.
    """

    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return ""
        return str(value)

    def _deserialize(self, value, attr, data, **kwargs):
        try:
            return get_registry().Unit(value)
        except ValueError as error:
            raise ValidationError("Invalid Pint unit.") from error



import os
from pathlib import Path
from jinja2 import Environment, PackageLoader, select_autoescape
from avlos.generators.filters import (
    avlos_enum_eps,
    avlos_bitmask_eps,
    file_from_path,
    capitalize_first,
)

env = Environment(loader=PackageLoader("avlos"), autoescape=select_autoescape())


def process(instance, config):
    env.filters["enum_eps"] = avlos_enum_eps
    env.filters["bitmask_eps"] = avlos_bitmask_eps
    env.filters["file_from_path"] = file_from_path
    env.filters["capitalize_first"] = capitalize_first
    process_helpers(instance, config)
    process_header(instance, config)
    process_impl(instance, config)


def process_helpers(instance, config):
    template = env.get_template("tm_helpers.hpp.jinja")
    file_path = config["paths"]["output_helpers"]
    os.makedirs(os.path.dirname(config["paths"]["output_helpers"]), exist_ok=True)
    with open(file_path, "w") as output_file:
        print(
            template.render(instance=instance),
            file=output_file,
        )


def process_header(instance, config):
    template = env.get_template("device.hpp.jinja")
    file_path = config["paths"]["output_header"]
    helper_file = config["paths"]["output_helpers"]
    try:
        includes = config["header_includes"]
    except KeyError:
        includes = []
    os.makedirs(os.path.dirname(config["paths"]["output_header"]), exist_ok=True)
    with open(file_path, "w") as output_file:
        print(
            template.render(
                instance=instance,
                includes=includes,
                helper_file=helper_file,
                device_name=Path(config["paths"]["output_header"]).stem,
            ),
            file=output_file,
        )
    for attr in instance.remote_attributes.values():
        if hasattr(attr, "remote_attributes"):
            recurse_header(attr, config)


def recurse_header(remote_object, config):
    template = env.get_template("remote_object.hpp.jinja")
    file_path = os.path.join(
        os.path.dirname(config["paths"]["output_header"]),
        remote_object.name + ".hpp",
    )
    helper_file = config["paths"]["output_helpers"]
    os.makedirs(os.path.dirname(config["paths"]["output_header"]), exist_ok=True)
    with open(file_path, "w") as output_file:
        print(
            template.render(instance=remote_object, helper_file=helper_file),
            file=output_file,
        )
    for attr in remote_object.remote_attributes.values():
        if hasattr(attr, "remote_attributes"):
            recurse_header(attr, config)


def process_impl(instance, config):
    template = env.get_template("device.cpp.jinja")
    file_path = config["paths"]["output_impl"]
    try:
        includes = config["impl_includes"]
    except KeyError:
        includes = []
    includes.append(Path(config["paths"]["output_header"]).name)
    os.makedirs(os.path.dirname(config["paths"]["output_impl"]), exist_ok=True)
    with open(file_path, "w") as output_file:
        print(
            template.render(
                instance=instance,
                includes=includes,
                device_name=Path(config["paths"]["output_header"]).stem,
            ),
            file=output_file,
        )
    for attr in instance.remote_attributes.values():
        if hasattr(attr, "remote_attributes"):
            recurse_impl(attr, config)


def recurse_impl(remote_object, config):
    template = env.get_template("remote_object.cpp.jinja")
    file_path = os.path.join(
        os.path.dirname(config["paths"]["output_impl"]),
        remote_object.name + ".cpp",
    )
    os.makedirs(os.path.dirname(config["paths"]["output_impl"]), exist_ok=True)
    with open(file_path, "w") as output_file:
        print(template.render(instance=remote_object), file=output_file)
    for attr in remote_object.remote_attributes.values():
        if hasattr(attr, "remote_attributes"):
            recurse_impl(attr, config)



class DummyChannel:
    """
    Dummy channel class
    """

    def __init__(self, value=0):
        self.value = value
        self.write = False

    def send(self, data, ep_id):
        if self.write:
            self.value = data

    def recv(self, ep_id):
        return [self.value]

    def set_value(self, value):
        self.value = value

    def write_on(self):
        self.write = True

    def write_off(self):
        self.write = False

    @property
    def serializer(self):
        return DummyCodec()


class DummyCodec:
    """
    Dummy CODEC class
    """

    def serialize(self, values, *args):
        return values[0]

    def deserialize(self, data, *args):
        return data



from marshmallow import (
    Schema,
    fields,
    post_load,
)
from avlos.unit_field import UnitField
from avlos.datatypes import DataTypeField
from avlos.mixins.comm_node import CommNode
from avlos.mixins.named_node import NamedNode
from avlos.mixins.meta_node import MetaNode
from avlos.mixins.func_attr_node import FuncAttrNode


class RemoteFunction(CommNode, NamedNode, MetaNode):
    """
    Remote Function with zero or more arguments, return
    type, parent and a comms channel
    """

    def __init__(
        self,
        name,
        summary,
        caller_name,
        arguments,
        func_attr=None,
        dtype=None,
        unit=None,
        rst_target=None,
        ep_id=-1,
        meta={},
    ):
        CommNode.__init__(self)
        NamedNode.__init__(self, name)
        MetaNode.__init__(self, meta_dict=meta)
        FuncAttrNode.__init__(self, func_attr)
        self.summary = summary
        self.dtype = dtype
        self.unit = unit
        self.caller_name = caller_name
        self.arguments = arguments
        self.rst_target = rst_target
        self.ep_id = ep_id

    def __call__(self, *args):
        mags = []
        for arg_val, arg_obj in zip(args, self.arguments):
            try:
                mags.append(arg_val.to(arg_obj.unit).magnitude)
            except AttributeError:
                mags.append(arg_val)
        data = self.channel.serializer.serialize(
            mags, *[arg.dtype for arg in self.arguments]
        )
        self.channel.send(data, self.ep_id)
        if not self.dtype.is_void:
            data = self.channel.recv(self.ep_id)
            value, *_ = self.channel.serializer.deserialize(data, self.dtype)
            try:
                return value * self.unit
            except TypeError:
                pass
            return value

    def str_dump(self):
        return "{}({}) -> {}".format(
            self.name,
            ", ".join([arg.as_function_argument for arg in self.arguments]),
            self.dtype.nickname,
        )


class RemoteArgument:
    """
    Class representing a RemoteFunction argument
    """

    def __init__(self, name, dtype, unit=None, summary=None):
        self.name = name
        self.dtype = dtype
        self.unit = unit
        self.summary = summary

    @property
    def as_function_argument(self):
        return " ".join([self.dtype.nickname, self.name])


class RemoteArgumentSchema(Schema):
    """
    Custom Marshmallow schema for generating RemoteFunction
    arguments
    """

    name = fields.String(
        required=True, error_messages={"required": "Name is required."}
    )
    summary = fields.String()
    dtype = DataTypeField(required=True)
    unit = UnitField()

    @post_load
    def make_remote_argument(self, data, **kwargs):
        return RemoteArgument(**data)



from avlos.counter import get_counter, make_counter, delete_counter

import unittest


class TestCounter(unittest.TestCase):
    def test_make_counter_return(self):
        self.assertIsNone(make_counter())

    def test_get_counter(self):
        delete_counter()
        self.assertIsNone(get_counter())
        make_counter()
        self.assertIsNotNone(get_counter())

    def test_make_counter(self):
        make_counter()
        counter1 = get_counter()
        counter2 = get_counter()
        make_counter()
        counter3 = make_counter()
        self.assertEqual(counter1, counter2)
        self.assertNotEqual(counter1, counter3)

    def test_counter_next(self):
        make_counter()
        counter = get_counter()
        self.assertEqual(0, counter.count)
        self.assertEqual(0, counter.next())
        self.assertEqual(1, counter.next())
        self.assertEqual(counter.next() + 1, counter.count)



from functools import cached_property
from marshmallow import fields, post_load
from avlos.definitions import RemoteNode, RemoteNodeSchema


class RootNode(RemoteNode):
    """
    Represent a remote root node with additional attributes.
    Extend from the RemoteNode class.
    """

    def __init__(self, version=None, *args, **kwargs):
        """
        Initialize a new instance of the RootNode.

        Args:
            version (str, optional): Version of the root node. Deprecated, defaults to None.
            *args: Variable length argument list to be passed to the superclass.
            **kwargs: Arbitrary keyword arguments to be passed to the superclass.
        """
        super().__init__(*args, **kwargs)

    @cached_property
    def root(self):
        """
        Cached property that returns the root node itself.

        Returns:
            RootNode: Self instance representing the root node.
        """
        return self


class RootNodeSchema(RemoteNodeSchema):
    """
    Custom Marshmallow schema for the root node
    """

    # The version field is deprecated since 0.6.0, but maintained
    # here to avoid breaking specs
    version = fields.Str()

    @post_load
    def make_remote_node(self, data, **kwargs):
        if "remote_attributes" in data:
            node = RootNode(**data)
            for child in node.remote_attributes.values():
                child._parent = node
            return node



class Counter:
    """
    A plain counter class with the ability
    to return the current and next index with
    auto-increment
    """

    def __init__(self, starting=0):
        self.count = starting

    def next(self):
        cnt = self.count
        self.count += 1
        return cnt

    def count(self):
        return self.count


counter = None


def make_counter():
    """
    Create a new counter and assign it to the
    global reference
    """
    global counter
    counter = Counter()


def get_counter():
    """
    Get the global counter
    """
    return counter


def delete_counter():
    """
    Delete the global counter
    """
    global counter
    counter = None



class ImpexNode:
    def import_values(self, data):
        # TODO: Assert correct type
        try:
            if "export" in self.meta and self.meta["export"] == True:
                self.set_value_with_string(data)
        except AttributeError:
            for name, attr in self.remote_attributes.items():
                try:
                    if name in data:
                        attr.import_values(data[name])
                except AttributeError:
                    pass

    def export_values(self):
        try:
            if "export" in self.meta and self.meta["export"] == True:
                return self.get_value()
        except AttributeError:
            values = {}
            for name, node in self.remote_attributes.items():
                try:
                    val = node.export_values()
                    if val != None:
                        values[name] = val
                except AttributeError:
                    pass
            if len(values) > 0:
                return values
            return None



import os
from copy import copy


def avlos_endpoints(input):
    """
    Traverse remote dictionary and return list
    of remote endpoints
    """
    def traverse_endpoint_list(ep_list, ep_out_list):
        for ep in ep_list:
            if (
                hasattr(ep, "getter_name")
                or hasattr(ep, "setter_name")
                or hasattr(ep, "caller_name")
            ):
                ep_out_list.append(ep)
            elif hasattr(ep, "remote_attributes"):
                traverse_endpoint_list(ep.remote_attributes.values(), ep_out_list)

    ep_out_list = []
    if hasattr(input, "remote_attributes"):
        traverse_endpoint_list(input.remote_attributes.values(), ep_out_list)
    return ep_out_list


def avlos_enum_eps(input):
    """
    Traverse remote dictionary and return a list of enum type endpoints
    """
    return [ep for ep in avlos_endpoints(input) if hasattr(ep, "options")]


def avlos_bitmask_eps(input):
    """
    Traverse remote dictionary and return a list of bitmask type endpoints
    """
    return [ep for ep in avlos_endpoints(input) if hasattr(ep, "bitmask")]


def as_include(input):
    """
    Render a string as a C include, with opening
    and closing braces or quotation marks
    """
    if input.startswith('"') and input.endswith('"'):
        return input
    elif input.startswith("<") and input.endswith(">"):
        return input
    return "<" + input + ">"


def file_from_path(input):
    """
    Get the file string from a path string
    """
    return os.path.basename(input)


def capitalize_first(input):
    """
    Capitalize the first character of a
    string, leaving the rest unchanged
    """
    return input[0].upper() + input[1:]



import yaml
import subprocess
from pathlib import Path
import importlib.resources
from avlos.deserializer import deserialize
from avlos.processor import process_with_config_file
import avlos.generators.generator_c as generator_c
import avlos.generators.generator_cpp as generator_cpp
import avlos.generators.generator_rst as generator_rst
from rstcheck_core import _extras, config as config_mod, runner
import unittest


class TestGeneration(unittest.TestCase):
    def test_c_output_manual(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath("definition/good_device.yaml")
        )
        enum_path_str = str(
            importlib.resources.files("tests").joinpath("outputs/tm_enums.h")
        )
        header_path_str = str(
            importlib.resources.files("tests").joinpath("outputs/test.h")
        )
        impl_path_str = str(
            importlib.resources.files("tests").joinpath("outputs/test.c")
        )
        with open(def_path_str) as device_desc_stream:
            obj = deserialize(yaml.safe_load(device_desc_stream))
            config = {
                "hash_string": "0x9e8dc7ac",
                "paths": {
                    "output_enums": enum_path_str,
                    "output_header": header_path_str,
                    "output_impl": impl_path_str,
                },
                "c_includes": {"src/common.h", "src/tm_enums.h"},
            }
            generator_c.process(obj, config)

            result = subprocess.run(
                ["cppcheck", "--error-exitcode=1", header_path_str, impl_path_str],
                stdout=subprocess.DEVNULL,
            )
            self.assertEqual(result.returncode, 0)

    def test_cpp_output_manual(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath("definition/good_device.yaml")
        )
        helper_path_str = str(
            importlib.resources.files("tests").joinpath("outputs/tm_helpers.hpp")
        )
        header_path_str = str(
            importlib.resources.files("tests").joinpath("outputs/base_device.hpp")
        )
        impl_path_str = str(
            importlib.resources.files("tests").joinpath("outputs/base_device.cpp")
        )
        with open(def_path_str) as device_desc_stream:
            obj = deserialize(yaml.safe_load(device_desc_stream))
            config = {
                "hash_string": "0x9e8dc7ac",
                "paths": {
                    "output_helpers": helper_path_str,
                    "output_header": header_path_str,
                    "output_impl": impl_path_str,
                },
                "header_includes": {"string"},
            }
            generator_cpp.process(obj, config)

            result = subprocess.run(
                ["cppcheck", "--error-exitcode=1", header_path_str, impl_path_str],
                stdout=subprocess.DEVNULL,
            )
            self.assertEqual(result.returncode, 0)

    def test_rst_output_manual(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath("definition/good_device.yaml")
        )
        out_path_str = str(
            importlib.resources.files("tests").joinpath("outputs/test.rst")
        )
        with open(def_path_str) as device_desc_stream:
            obj = deserialize(yaml.safe_load(device_desc_stream))
            config = {
                "hash_string": "0x9e8dc7ac",
                "paths": {"output_file": out_path_str},
            }
            generator_rst.process(obj, config)

        rstcheck_config = config_mod.RstcheckConfig()
        path = Path(out_path_str)
        _runner = runner.RstcheckMainRunner(
            check_paths=[path], rstcheck_config=rstcheck_config, overwrite_config=False
        )
        _runner.check()
        _runner.print_result()

    def test_avlos_config(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath("definition/good_device.yaml")
        )
        config_file_path_str = str(
            importlib.resources.files("tests").joinpath("definition/avlos_config.yaml")
        )
        with open(def_path_str) as device_desc_stream:
            obj = deserialize(yaml.safe_load(device_desc_stream))
            process_with_config_file(obj, config_file_path_str)



from avlos.mixins.comm_node import CommNode
from avlos.mixins.named_node import NamedNode
from avlos.mixins.meta_node import MetaNode
from avlos.mixins.impex_node import ImpexNode
from avlos.datatypes import DataType
from avlos.mixins.func_attr_node import FuncAttrNode


class RemoteEnum(CommNode, NamedNode, MetaNode, ImpexNode):
    """
    Remote Endpoint with a value represented as an enum
    """

    enum_type = DataType.UINT8

    def __init__(
        self,
        name,
        summary,
        func_attr=None,
        getter_name=None,
        setter_name=None,
        options=None,
        rst_target=None,
        export=False,
        ep_id=-1,
        meta={},
    ):
        CommNode.__init__(self)
        NamedNode.__init__(self, name)
        MetaNode.__init__(self, meta_dict=meta)
        FuncAttrNode.__init__(self, func_attr)
        self.summary = summary
        self.options = options  # options is needed to deserialize
        self.getter_name = getter_name
        self.setter_name = setter_name
        self.export = export
        self.rst_target = rst_target
        self.ep_id = ep_id

    def get_value(self):
        """
        Retrieve the current enum value from the remote endpoint.
        Sends a request, receives the raw integer value, and converts it to the corresponding enum member.

        Returns:
            The enum member corresponding to the remote value
        """
        assert self.getter_name
        self.channel.send([], self.ep_id)
        data = self.channel.recv(self.ep_id)
        value, *_ = self.channel.serializer.deserialize(data, self.enum_type)
        return self.options(value)

    def set_value(self, __value):
        """
        Set a new enum value on the remote endpoint.
        Accepts values as integers, enum members, or string names, and converts them appropriately.

        Args:
            __value: The value to set (int, enum member, or string name)

        Raises:
            ValueError: If the value cannot be converted to a valid enum member
        """
        assert self.setter_name
        # Check if the value is already an integer and within the range of the enum
        if isinstance(__value, int) and __value in self.options._value2member_map_:
            value = __value

        # Check if the value is a member of the enum
        elif isinstance(__value, self.options):
            value = __value.value

        # Check if the value is a string corresponding to an enum member
        elif isinstance(__value, str) and __value in self.options._member_map_:
            value = self.options[__value].value

        else:
            raise ValueError(
                f"Invalid value: {__value}. Expected an integer, an enum member, or a string corresponding to an enum member."
            )
        data = self.channel.serializer.serialize([value], self.dtype)
        self.channel.send(data, self.ep_id)

    def export_options(self, namespace):
        """
        Export the members of the enum to the
        indicated namespace
        """
        if self.export:
            namespace.update(self.options.__members__)

    @property
    def dtype(self):
        """
        Mocks the endpoint datatype
        """
        return self.enum_type

    @property
    def unit(self):
        """
        Mocks the endpoint unit
        """
        return None

    def str_dump(self):
        """
        Generate a formatted string representation of the enum attribute and its current value.

        Returns:
            A formatted string showing the attribute name and current enum member
        """
        val = self.get_value()
        return "{0}: {1}".format(self.name, str(val))



import yaml
import importlib.resources
import urllib.request
from avlos.deserializer import deserialize
from avlos.definitions.remote_node import RemoteNodeSchema
import marshmallow
import pint
import unittest
from tests.dummy_channel import DummyChannel


class TestDeserialization(unittest.TestCase):
    def test_success(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath("definition/good_device.yaml")
        )
        with open(def_path_str) as device_description:
            obj = deserialize(yaml.safe_load(device_description))
            obj._channel = DummyChannel()
            print(obj)

    def test_success_url(self):
        device_desc_string = urllib.request.urlopen(
            "https://raw.githubusercontent.com/tinymovr/avlos/main/tests/definition/good_device.yaml"
        ).read()
        obj = deserialize(yaml.safe_load(device_desc_string))
        obj._channel = DummyChannel()
        print(obj)

    def test_undefined_unit(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath(
                "definition/bad_device_unit.yaml"
            )
        )
        with open(def_path_str) as device_description:
            with self.assertRaises(pint.errors.UndefinedUnitError):
                deserialize(yaml.safe_load(device_description))

    def test_bitmask_labels(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath("definition/good_device.yaml")
        )
        with open(def_path_str) as device_description:
            device = deserialize(yaml.safe_load(device_description))
            device._channel = DummyChannel()
            self.assertEqual(device.errors.value, 0)

    def test_empty_bitmask_labels(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath(
                "definition/bad_device_bitmask.yaml"
            )
        )
        with open(def_path_str) as device_description:
            with self.assertRaises(marshmallow.exceptions.ValidationError):
                deserialize(yaml.safe_load(device_description))

    def test_version_field_present(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath(
                "definition/obsolete_device.yaml"
            )
        )
        with open(def_path_str) as device_description:
            device = deserialize(yaml.safe_load(device_description))
            device._channel = DummyChannel()
            self.assertEqual(device.errors.value, 0)

    def test_validation_fail(self):
        def_path_str = str(
            importlib.resources.files("tests").joinpath(
                "definition/bad_device_name.yaml"
            )
        )
        with open(def_path_str) as device_description:
            with self.assertRaises(marshmallow.exceptions.ValidationError):
                deserialize(yaml.safe_load(device_description))

    def test_schema_field_initialization(self):
        """
        Test that RemoteNodeSchema can be instantiated without TypeError.
        This catches issues with using 'default' instead of 'dump_default' or 'load_default'
        in marshmallow 3.x field definitions, which causes:
        TypeError: Field.__init__() got an unexpected keyword argument 'default'
        """
        # This should not raise TypeError during schema instantiation
        schema = RemoteNodeSchema()

        # Test that schema can dump data with default values
        data = {"name": "test_node", "remote_attributes": []}
        result = schema.dump(data)

        # Verify the schema is working correctly
        self.assertIsInstance(result, dict)
        self.assertEqual(result["name"], "test_node")

        # Test that optional fields with dump_default are included
        self.assertIn("func_attr", result)
        self.assertIsNone(result["func_attr"])
        self.assertIn("export", result)
        self.assertEqual(result["export"], False)
        self.assertIn("ep_id", result)
        self.assertEqual(result["ep_id"], -1)



"""Avlos CLI

Usage:
    avlos from (file <spec_path> | url <spec_url>) [--config=<config_path>]
    avlos -h | --help
    avlos --version

Options:
    --config=<config_path>  Path of the Avlos config file [default: ./avlos_config.yaml]
"""

import yaml
from typing import Dict
import logging
import pkg_resources
import urllib.request
from docopt import docopt
from avlos.deserializer import deserialize
from avlos.processor import process_with_config_file

shell_name = "Avlos"


def run_cli():
    """
    Runs the Avlos CLI
    """
    version: str = pkg_resources.require("avlos")[0].version
    arguments: Dict[str, str] = docopt(__doc__, version=shell_name + " " + str(version))

    logger = configure_logging()

    config_path: str = arguments["--config"]

    if arguments["<spec_path>"]:
        with open(arguments["<spec_path>"]) as device_desc_stream:
            obj = deserialize(yaml.safe_load(device_desc_stream))
            process_with_config_file(obj, config_path, traverse_path=True)
    elif arguments["<spec_url>"]:
        device_desc_string = urllib.request.urlopen(arguments["<spec_url>"]).read()
        obj = deserialize(yaml.safe_load(device_desc_string))
        process_with_config_file(obj, config_path, traverse_path=True)


def configure_logging() -> logging.Logger:
    """
    Configures logging options and
    generates a default logger instance.
    """
    logger = logging.getLogger("avlos")
    logger.setLevel(logging.DEBUG)
    return logger



from functools import cached_property


class NamedNode:
    def __init__(self, name, include_base_name=False):
        self._parent = None
        self.name = name
        self.include_base_name = include_base_name

    # this kinda stinks, cause it is implemented elsewhere
    # @cached_property
    # def parent(self):
    #     return self._parent

    @cached_property
    def full_name(self):
        try:
            return "{}.{}".format(self._parent.full_name, self.name).strip(".")
        except AttributeError:
            if self.include_base_name:
                return self.name
            return ""



#!/usr/bin/env python

import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name="Avlos",
    use_scm_version={
        "write_to": "avlos/_version.py",
        "write_to_template": '__version__ = "{version}"\n',
    },
    description="Generate type-safe communication protocols for embedded systems from a single YAML specification",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Yannis Chatzikonstantinou",
    author_email="yannis@tinymovr.com",
    url="https://www.tinymovr.com",
    packages=find_packages(include=["avlos", "avlos.*"]),
    include_package_data=True,
    python_requires=">=3.9",
    setup_requires=["setuptools_scm"],
    install_requires=["marshmallow", "pyyaml", "pint", "docopt", "jinja2"],
    extras_require={"dev": ["rstcheck"]},
    entry_points={"console_scripts": ["avlos=avlos.cli:run_cli"]},
)



from os.path import join, dirname, basename, realpath
import yaml
from importlib import import_module


def process_with_config_file(device_instance, avlos_config_path, traverse_path=False):
    """
    Process a device spec using an output config path.
    """
    real_path = realpath(avlos_config_path)
    avlos_config = None
    while not avlos_config:
        try:
            with open(real_path) as avlos_config_stream:
                avlos_config = yaml.safe_load(avlos_config_stream)
        except FileNotFoundError as e:
            if traverse_path:
                # traverse path backwards
                new_config_path = join(dirname(real_path), "..", basename(real_path))
                if new_config_path == real_path:
                    # we're at the root, and have not found the file
                    raise FileNotFoundError from e
                real_path = new_config_path
            else:
                raise e

    for module_config in avlos_config["generators"].values():
        # Convert all paths to absolute using the base path
        for path_name, path in module_config["paths"].items():
            module_config["paths"][path_name] = realpath(join(dirname(real_path), path))
    process_with_config_object(device_instance, avlos_config)


def process_with_config_object(device_instance, avlos_config):
    """
    Process a device spec using an output config object.
    """
    for module_name, module_config in avlos_config["generators"].items():
        if "enabled" in module_config and True == module_config["enabled"]:
            generator = import_module(
                ".generators.{}".format(module_name), package="avlos"
            )
            generator.process(device_instance, module_config)



from avlos.mixins.comm_node import CommNode
from avlos.mixins.named_node import NamedNode
from avlos.mixins.meta_node import MetaNode
from avlos.mixins.impex_node import ImpexNode
from avlos.datatypes import DataType
from avlos.mixins.func_attr_node import FuncAttrNode


class RemoteBitmask(CommNode, NamedNode, MetaNode, ImpexNode):
    """
    Remote Endpoint with a value represented as a bitmask
    """

    flag_type = DataType.UINT8

    def __init__(
        self,
        name,
        summary,
        func_attr=None,
        getter_name=None,
        setter_name=None,
        flags=None,
        rst_target=None,
        export=False,
        ep_id=-1,
        meta={},
    ):
        CommNode.__init__(self)
        NamedNode.__init__(self, name)
        MetaNode.__init__(self, meta_dict=meta)
        FuncAttrNode.__init__(self, func_attr)
        self.summary = summary
        self.bitmask = flags  # flags is needed to deserialize
        self.getter_name = getter_name
        self.setter_name = setter_name
        self.export = export
        self.rst_target = rst_target
        self.ep_id = ep_id

    def get_value(self):
        """
        Retrieve the current bitmask value from the remote endpoint.
        Sends a request, receives the raw integer value, and converts it to a bitmask object.

        Returns:
            The bitmask object representing the active flags
        """
        assert self.getter_name
        self.channel.send([], self.ep_id)
        data = self.channel.recv(self.ep_id)
        value, *_ = self.channel.serializer.deserialize(data, self.flag_type)
        return self.bitmask(value)

    def set_value(self, __value):
        """
        Setting bitmask values is not currently supported.

        Raises:
            NotImplementedError: Always raised as this operation is not implemented
        """
        raise NotImplementedError

    def export_flags(self, namespace):
        """
        Export the members of the bitmask to the
        indicated namespace
        """
        if self.export:
            namespace.update(self.bitmask.__members__)

    @property
    def dtype(self):
        """
        Mocks the endpoint datatype
        """
        return self.flag_type

    @property
    def unit(self):
        """
        Mocks the endpoint unit
        """
        return None

    def str_dump(self):
        """
        Generate a formatted string representation of the bitmask attribute and its current value.

        Returns:
            A formatted string showing the attribute name and active flags, or indication of no flags
        """
        val = self.get_value()
        return "{0}: {1}".format(
            self.name,
            str(val) if val > 0 else "(no flags)",
        )



import os
from jinja2 import Environment, PackageLoader, select_autoescape
from avlos.generators.filters import (
    avlos_endpoints,
    avlos_enum_eps,
    avlos_bitmask_eps,
    as_include,
)

env = Environment(loader=PackageLoader("avlos"), autoescape=select_autoescape())


def process(instance, config):
    env.filters["endpoints"] = avlos_endpoints
    env.filters["enum_eps"] = avlos_enum_eps
    env.filters["bitmask_eps"] = avlos_bitmask_eps
    env.filters["as_include"] = as_include

    template = env.get_template("tm_enums.h.jinja")
    os.makedirs(os.path.dirname(config["paths"]["output_enums"]), exist_ok=True)
    with open(config["paths"]["output_enums"], "w") as output_file:
        print(
            template.render(instance=instance),
            file=output_file,
        )

    template = env.get_template("fw_endpoints.h.jinja")
    try:
        includes = config["header_includes"]
    except KeyError:
        includes = []
    os.makedirs(os.path.dirname(config["paths"]["output_header"]), exist_ok=True)
    with open(config["paths"]["output_header"], "w") as output_file:
        print(
            template.render(instance=instance, includes=includes),
            file=output_file,
        )

    template = env.get_template("fw_endpoints.c.jinja")
    try:
        includes = config["impl_includes"]
    except KeyError:
        includes = []
    os.makedirs(os.path.dirname(config["paths"]["output_impl"]), exist_ok=True)
    with open(config["paths"]["output_impl"], "w") as output_file:
        print(
            template.render(instance=instance, includes=includes),
            file=output_file,
        )
